#include "opendrive_road_node.h"
#include "opendrive_road.h"
#include "exceptions.h"
#include <math.h>

COpendriveRoadNode::COpendriveRoadNode()
{
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->pose.x=0.0;
  this->pose.y=0.0;
  this->pose.heading=0.0;
  this->parents.clear();
  this->links.clear();
  this->index=-1;
}

COpendriveRoadNode *COpendriveRoadNode::clone(link_up_ref_t &new_link_ref)
{
  COpendriveRoadNode *new_node;

  new_node=new COpendriveRoadNode();
  new_node->resolution=this->resolution;
  new_node->scale_factor=this->scale_factor;
  new_node->pose=this->pose;
  new_node->parents=this->parents;
  new_node->links.resize(this->links.size());
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(new_link_ref.find(this->links[i])!=new_link_ref.end())
      new_node->links[i]=new_link_ref[this->links[i]];
    else
    {
      new_node->links[i]=new COpendriveLink(*this->links[i]);
      new_link_ref[this->links[i]]=new_node->links[i];
    }
  }
  new_node->index=this->index;

  return new_node;
}

bool COpendriveRoadNode::add_link(COpendriveRoadNode *node,opendrive_mark_t mark,COpendriveRoadSegment *segment,COpendriveLane *lane)
{
  TOpendriveRoadNodeParent *parent;
  TOpendriveWorldPose node_pose;
  double start_curvature,end_curvature;
  COpendriveLink *new_link;

  if(this->has_link_with(*node) || node->has_link_with(*this))
    return false;
  new_link=new COpendriveLink();
  new_link->set_prev(this);
  new_link->set_next(node);
  new_link->set_resolution(this->resolution);
  new_link->set_scale_factor(this->scale_factor);
  new_link->set_road_mark(mark);
  new_link->set_parent_segment(segment);
  new_link->set_parent_lane(lane);
  // get the curvature
  node_pose=node->get_pose();
  parent=this->get_parent_with_lane(lane);
  if(parent!=NULL)
  {
    start_curvature=parent->start_curvature;
    end_curvature=parent->end_curvature;
  }
  else
  {
    if(this->get_num_parents()==1)
    {
      start_curvature=this->parents[0].start_curvature;
      end_curvature=this->parents[0].end_curvature;
    }
    else
    {
      start_curvature=0.0;
      end_curvature=0.0;
    }
  }
  new_link->generate(start_curvature,end_curvature);
  this->add_link(new_link);
  node->add_link(new_link);

  return true;
}

void COpendriveRoadNode::add_link(COpendriveLink *link)
{
  if(!this->has_link(link))
    this->links.push_back(link);
}

void COpendriveRoadNode::remove_link(COpendriveLink *link)
{
  std::vector<COpendriveLink *>::iterator it;

  for(it=this->links.begin();it!=this->links.end();)
  {
    if((*it)==link)
    {
      if((*it)->prev==this)
        delete (*it);
      it=this->links.erase(it);
      break;
    }
    else
      it++;
  }
}

bool COpendriveRoadNode::has_link(COpendriveLink *link)
{
  for(unsigned int i=0;i<this->links.size();i++)
    if(this->links[i]==link)
      return true;

  return false;
}

bool COpendriveRoadNode::has_link_with(const COpendriveRoadNode &node) const
{
  for(unsigned int i=0;i<this->links.size();i++)
    if(this->links[i]->prev==this && this->links[i]->next==&node)
      return true;

  return false;
}

bool COpendriveRoadNode::has_link_with(unsigned int node_index) const
{
  for(unsigned int i=0;i<this->links.size();i++)
    if(this->links[i]->prev==this && this->links[i]->next->get_index()==node_index)
      return true;

  return false;
}

COpendriveLink *COpendriveRoadNode::get_link_with(COpendriveRoadNode *node)
{
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(this->links[i]->prev==node || this->links[i]->next==node)
      return this->links[i];
  }

  return NULL;
}

void COpendriveRoadNode::set_resolution(double res)
{
  this->resolution=res;

  for(unsigned int i=0;i<this->links.size();i++)
    this->links[i]->set_resolution(res);
}

void COpendriveRoadNode::set_scale_factor(double scale)
{
  this->scale_factor=scale;

  for(unsigned int i=0;i<this->links.size();i++)
    this->links[i]->set_scale_factor(scale);

  for(unsigned int i=0;i<this->parents.size();i++)
  {
    this->parents[i].start_curvature*=scale/this->scale_factor;
    this->parents[i].end_curvature*=scale/this->scale_factor;
  }
}

void COpendriveRoadNode::set_index(unsigned int index)
{
  this->index=index;
}

void COpendriveRoadNode::set_pose(TOpendriveWorldPose &start)
{
  this->pose=start;
}

void COpendriveRoadNode::add_parent(COpendriveLane *lane,COpendriveRoadSegment *segment,double start_curvature,double end_curvature)
{
  TOpendriveRoadNodeParent new_parent;

  for(unsigned int i=0;i<this->parents.size();i++)
    if(this->parents[i].lane==lane && this->parents[i].segment==segment)
      return;
  new_parent.lane=lane;
  new_parent.segment=segment;
  new_parent.start_curvature=start_curvature;
  new_parent.end_curvature=end_curvature;
  this->parents.push_back(new_parent);
}

TOpendriveRoadNodeParent *COpendriveRoadNode::get_parent_with_lane(const COpendriveLane *lane)
{
  for(unsigned int i=0;i<this->parents.size();i++)
    if(this->parents[i].lane==lane)
      return &this->parents[i];

  return NULL;
}

TOpendriveRoadNodeParent *COpendriveRoadNode::get_parent_with_segment(const COpendriveRoadSegment *segment)
{
  for(unsigned int i=0;i<this->parents.size();i++)
    if(this->parents[i].segment==segment)
      return &this->parents[i];

  return NULL;
}


void COpendriveRoadNode::free(void)
{
  this->parents.clear();
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(this->links[i]->prev==this)
    {
      delete this->links[i];
      this->links[i]=NULL;
    }
  }
  this->links.clear();
}

bool COpendriveRoadNode::updated(node_up_ref_t &refs)
{
  node_up_ref_t::iterator updated_it;

  for(updated_it=refs.begin();updated_it!=refs.end();updated_it++)
    if(updated_it->second==this)
      return true;

  return false;
}

COpendriveRoadNode *COpendriveRoadNode::get_original_node(node_up_ref_t &node_refs)
{
  node_up_ref_t::iterator updated_it;

  for(updated_it=node_refs.begin();updated_it!=node_refs.end();updated_it++)
    if(updated_it->second==this)
      return updated_it->first;

  return NULL;
}

void COpendriveRoadNode::update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs)
{
  std::vector<COpendriveLink *>::iterator link_it;

  if(this->updated(node_refs))
  {
    for(link_it=this->links.begin();link_it!=this->links.end();link_it++)
      (*link_it)->update_references(segment_refs,lane_refs,node_refs);
    for(unsigned int i=0;i<this->parents.size();i++)
    {
      if(segment_refs.find(this->parents[i].segment)!=segment_refs.end())
        this->parents[i].segment=segment_refs[this->parents[i].segment];
      if(lane_refs.find(this->parents[i].lane)!=lane_refs.end())
        this->parents[i].lane=lane_refs[this->parents[i].lane];
    }
  }
}

void COpendriveRoadNode::clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs)
{
  std::vector<TOpendriveRoadNodeParent>::iterator parent_it;
  std::vector<COpendriveLink *>::iterator link_it;

  if(this->updated(node_refs))
  {
    for(link_it=this->links.begin();link_it!=this->links.end();)
    {
      if((*link_it)->clean_references(node_refs))
        link_it=this->links.erase(link_it);
      else
        link_it++;
    }
    for(parent_it=this->parents.begin();parent_it!=this->parents.end();)
    {
      if(!parent_it->segment->updated(segment_refs) || !parent_it->lane->updated(lane_refs))
        parent_it=this->parents.erase(parent_it);
      else
        parent_it++;
    }
  }
}

void COpendriveRoadNode::update_start_pose(COpendriveLane *lane,TOpendriveWorldPose &start,double distance)
{
  std::vector<TOpendriveRoadNodeParent>::iterator parent_it;
  std::vector<COpendriveLink *>::iterator link_it;
  TOpendriveRoadNodeParent *parent;

  this->pose=start;
  // remove the references to all lanes and segments except for lane
  for(parent_it=this->parents.begin();parent_it!=this->parents.end();)
  {
    if(parent_it->lane!=lane)
      parent_it=this->parents.erase(parent_it);
    else
    {
      parent=&(*parent_it);
      parent_it++;
    }
  }
  // update the links
  for(link_it=this->links.begin();link_it!=this->links.end();)
  {
    if((*link_it)->next==this)
    {
      delete *link_it;
      link_it=this->links.erase(link_it); 
    }
    else
    {
      if((*link_it)->lane==lane)
      {
        parent->start_curvature=(*link_it)->get_curvature_at(distance);
        (*link_it)->update_start_pose(start,parent->start_curvature);
      }
      else
        (*link_it)->update_start_pose(start,0.0);
      link_it++;
    }
  }
}

void COpendriveRoadNode::update_end_pose(COpendriveLane *lane,TOpendriveWorldPose &end,double distance)
{
  std::vector<TOpendriveRoadNodeParent>::iterator parent_it;
  std::vector<COpendriveLink *>::iterator link_it;

  // remove the references to all lanes and segments except for lane
  for(parent_it=this->parents.begin();parent_it!=this->parents.end();)
  { 
    if(parent_it->lane!=lane)
      parent_it=this->parents.erase(parent_it);
    else
      parent_it++;
  }
  // update the links
  for(link_it=this->links.begin();link_it!=this->links.end();)
  {
    if((*link_it)->prev==this)
    {
      delete *link_it;
      link_it=this->links.erase(link_it);
    }
    else
      link_it++;
  }
}

double COpendriveRoadNode::get_resolution(void) const
{
  return this->resolution;
}

double COpendriveRoadNode::get_scale_factor(void) const
{
  return this->scale_factor;
}

unsigned int COpendriveRoadNode::get_index(void) const
{
 return this->index;
}

unsigned int COpendriveRoadNode::get_num_parents(void) const
{
  return this->parents.size();
}

const COpendriveRoadSegment &COpendriveRoadNode::get_parent_segment(unsigned int index) const
{
  std::stringstream text;

  if(index>=0 && index<this->parents.size())
    return *this->parents[index].segment;
  else
  {
    text << "Invalid parent index " << index << " for node " << this->index << " (has " << this->parents.size() << " parents)";
    throw CException(_HERE_,text.str());
  }
}

const COpendriveLane &COpendriveRoadNode::get_parent_lane(unsigned int index) const
{
  std::stringstream text;

  if(index>=0 && index<this->parents.size())
    return *this->parents[index].lane;
  else
  {
    text << "Invalid parent index " << index << " for node " << this->index << " (has " << this->parents.size() << " parents)";
    throw CException(_HERE_,text.str());
  }
}

TOpendriveWorldPose COpendriveRoadNode::get_pose(void) const
{
  return this->pose;
}

unsigned int COpendriveRoadNode::get_num_links(void) const
{
  return this->links.size();
}

const COpendriveLink &COpendriveRoadNode::get_link(unsigned int index) const
{
  std::stringstream text;

  if(index>=0 && index<this->links.size())
    return *this->links[index];
  else
  {
    text << "Invalid link index " << index << " for node " << this->index << " (has " << this->links.size() << " links)";
    throw CException(_HERE_,text.str());
  }
}

const COpendriveLink &COpendriveRoadNode::get_link_ending_at(unsigned int node_index) const
{
  std::stringstream text;

  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(this->links[i]->get_next().get_index()==node_index)
      return *this->links[i];
  }

  text << "No link in node " << this->index << " ends at node " << node_index;
  throw CException(_HERE_,text.str());
}

const COpendriveLink &COpendriveRoadNode::get_link_ending_at(const COpendriveRoadNode &node) const
{
  std::stringstream text;

  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_next()==&node)
      return *this->links[i];
  }

  text << "No link in node " << this->index << " ends at node " << node.get_index();
  throw CException(_HERE_,text.str());
}

unsigned int COpendriveRoadNode::get_closest_link_index(TOpendriveWorldPose &pose,double &distance,bool only_lanes,double angle_tol) const 
{
  double dist,min_dist=std::numeric_limits<double>::max();
  unsigned int closest_index=-1;
  double angle,length,min_length=std::numeric_limits<double>::max();
  TOpendriveWorldPose closest_pose;

  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      if(!only_lanes || (only_lanes && this->links[i]->lane!=NULL))
      {
        length=this->links[i]->find_closest_world_pose(pose,closest_pose);
        if(length<std::numeric_limits<double>::max())
        {
          angle=diff_angle(normalize_angle(pose.heading),closest_pose.heading);
          if(fabs(angle)<angle_tol)
          {
            dist=sqrt(pow(closest_pose.x-pose.x,2.0)+pow(closest_pose.y-pose.y,2.0));
            if(dist<min_dist)
            {
              min_dist=dist;
              min_length=length;
              closest_index=i;
            }
          }
        }
      }
    }
  }

  distance=min_length;
  return closest_index;
}

const CopendriveLink &COpendriveRoadNode::get_closest_link(TOpendriveWorldPose &pose,double &distance,bool only_lanes,double angle_tol)const
{
  unsigned int closest_index;

  closest_index=this->get_closest_link_index(pose,distance,only_lanes,angle_tol);
  if(closest_index==(unsigned int)-1)
    throw CException(_HERE_,"Impossible to find the closest link");
  return *this->links[closest_index];
}

double COpendriveRoadNode::get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,bool only_lanes,double angle_tol) const
{
  double dist,min_dist=std::numeric_limits<double>::max();
  double angle;
  double length,closest_length=std::numeric_limits<double>::max();
  TOpendriveWorldPose tmp_pose;

  closest_pose.x=std::numeric_limits<double>::max();
  closest_pose.y=std::numeric_limits<double>::max();
  closest_pose.heading=std::numeric_limits<double>::max();
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      if(!only_lanes || (only_lanes && this->links[i]->lane!=NULL))
      {
        length=this->links[i]->find_closest_world_pose(pose,tmp_pose);
        if(length<std::numeric_limits<double>::max())
        {
          angle=diff_angle(normalize_angle(pose.heading),tmp_pose.heading);
          if(fabs(angle)<angle_tol)
          {
            dist=sqrt(pow(tmp_pose.x-pose.x,2.0)+pow(tmp_pose.y-pose.y,2.0));
            if(dist<min_dist)
            {
              min_dist=dist;
              closest_length=length;
              closest_pose=tmp_pose;
            }
          }
        }
      }
    }
  }

  return closest_length;
}

void COpendriveRoadNode::get_closest_poses(TOpendriveWorldPose &pose,std::vector<TOpendriveWorldPose> &closest_poses,std::vector<double> &distances,double dist_tol,bool only_lanes,double angle_tol) const
{
  double dist;
  double angle;
  double length;
  TOpendriveWorldPose closest_pose;
  bool already_added;

  closest_poses.clear();
  distances.clear();
  for(unsigned int i=0;i<this->links.size();i++)
  {
    if(&this->links[i]->get_prev()==this)// links starting at this node
    {
      if(!only_lanes || (only_lanes && this->links[i]->lane!=NULL))
      {
        length=this->links[i]->find_closest_world_pose(pose,closest_pose);
        if(length<std::numeric_limits<double>::max())
        {
          angle=diff_angle(normalize_angle(pose.heading),closest_pose.heading);
          if(fabs(angle)<angle_tol)
          {
            dist=sqrt(pow(closest_pose.x-pose.x,2.0)+pow(closest_pose.y-pose.y,2.0));
            if(dist<=dist_tol)
            {
              already_added=false;
              for(unsigned int j=0;j<closest_poses.size();j++)
                if(fabs(closest_poses[j].x-closest_pose.x)<this->resolution && 
                   fabs(closest_poses[j].y-closest_pose.y)<this->resolution)
                {
                  already_added=true;
                  break;
                }
              if(!already_added)
              {
                closest_poses.push_back(closest_pose);
                distances.push_back(length);
              }
            }
          }
        }
      }
    }
  }
}

std::ostream& operator<<(std::ostream& out, COpendriveRoadNode &node)
{
  out << "    Node: " << node.get_index() << std::endl;
  out << "      start pose: x: " << node.pose.x << " y: " << node.pose.y << " heading: " << node.pose.heading << std::endl;
  out << "      Number of parent: " << node.get_num_parents() << std::endl;
  for(unsigned int i=0;i<node.get_num_parents();i++)
  {
    out << "      Parent " << i << ":" << std::endl;
    out << "        segment " << node.get_parent_segment(i).get_name() << " (lane " << node.get_parent_lane(i).get_id() << ")" << std::endl; 
  }
  out << "      Number of links: " << node.links.size() << std::endl;
  for(unsigned int i=0;i<node.links.size();i++)
  {
    out << "      Link " << i << ":" << std::endl;
    out << *node.links[i];
  }
  return out;
}

COpendriveRoadNode::~COpendriveRoadNode()
{
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->pose.x=0.0;
  this->pose.y=0.0;
  this->pose.heading=0.0;
  this->free();
}

